# -*- coding: utf-8 -*-
'''
National Instruments Systems Management Beacon to react to the http connection change
'''
from __future__ import absolute_import

# Import Python libs
import logging
import os.path
import sys
import requests
from typing import Any, Dict

# Import salt libs
# pylint: disable=import-error,3rd-party-module-not-gated
import salt.utils.event
import salt.utils.platform
import salt.ext.six as six
# pylint: enable=import-error,3rd-party-module-not-gated

# pylint: disable=import-error,3rd-party-local-module-not-gated
if salt.utils.platform.is_windows():
    import salt.utils.winapi  # pylint: disable=ungrouped-imports
else:
    import salt.modules.cmdmod  # pylint: disable=ungrouped-imports
# pylint: enable=import-error,3rd-party-local-module-not-gated

log = logging.getLogger(__name__)

# Placeholder for 'modules/_nisysmgmt_utils.py' module that will be loaded
# in __virtual__ (when '__opts__' are available).
_nisysmgmt_utils = None  # pylint: disable=invalid-name

__virtualname__ = 'nisysmgmt_status_checks'

EVENT = None
BEACON_INTERVAL_WITH_HTTP_MASTER_FILE = 300
BEACON_INTERVAL_WITH_NO_HTTP_MASTER_FILE = 3
LAST_KNOWN_HTTP_CONNECTION_STATE = None
HTTP_MASTER_FILE_WAS_FOUND = False
STATUS_CHECKS = {}


def __virtual__():
    global _nisysmgmt_utils  # pylint: disable=global-statement,invalid-name
    # Add extra search paths to load '_nisysapi_ipc'
    paths_to_add = __opts__['module_dirs']
    extmods_module_dir = os.path.join(__opts__['extension_modules'], 'modules')
    if extmods_module_dir not in paths_to_add:
        paths_to_add.append(extmods_module_dir)
    sys.path.extend(paths_to_add)
    try:
        import _nisysmgmt_utils  # pylint: disable=redefined-outer-name,import-error,3rd-party-local-module-not-gated
    finally:
        # Remove the extra search paths that we added to
        # sys.path
        for path in paths_to_add:
            sys.path.remove(path)

    return __virtualname__


def _init_event():
    '''
    Initialize the event object if needed.
    '''
    global EVENT  # pylint: disable=global-statement

    if not EVENT:
        EVENT = salt.utils.event.get_event(
            'minion', opts=__opts__, listen=False
        )


def _fire_local_event(event_data: Dict[str, Any], event_tag: str) -> None:
    '''
    Fire an event to the local event bus.

    :param event_data: The data associated with the event.
    :param event_tag: The identifier of the event.
    '''
    global EVENT  # pylint: disable=global-statement

    if not EVENT:
        EVENT = salt.utils.event.get_event(
            'minion', opts=__opts__, listen=False
        )
    EVENT.fire_event(event_data, event_tag)


def _fire_status_change_event():
    '''
    Fire an event to update the values in STATUS_CHECKS
    '''
    _init_event()
    tag = 'nisysmgmt/status/update/{0}'.format(__grains__['id'])  # pylint: disable=undefined-variable
    data = {'data': STATUS_CHECKS}  # pylint: disable=undefined-variable
    EVENT.fire_master(data, tag)


def _minion_http_connected_status_changed():
    '''
    Check if the ``http_connected`` status has changed.

    :return: ``True`` if the ``http_connected`` status has changed. ``False`` otherwise.
    :rtype: bool
    '''

    global STATUS_CHECKS
    last_known_http_connection = _get_last_known_http_connection_state()
    http_connected = _check_http_connection()


    if last_known_http_connection != http_connected:
        STATUS_CHECKS['http_connected'] = http_connected  # pylint: disable=undefined-variable
        _set_last_known_http_connection_state(http_connected)
        return True
    return False


def _set_last_known_http_connection_state(is_connected):
    '''
    This will persist the last_known_http_connection_state
    to synchronize the beacon with set_http_connection_state/unset_http_connection_state
    functions in the module - to avoid race-condition on the master.
    '''
    global LAST_KNOWN_HTTP_CONNECTION_STATE  # pylint: disable=global-statement
    LAST_KNOWN_HTTP_CONNECTION_STATE = is_connected


def _get_last_known_http_connection_state():
    '''
    Return the persisted last_known_http_connection_state
    of the minion.
    '''
    return LAST_KNOWN_HTTP_CONNECTION_STATE


def _check_http_connection():
    '''
    Check if a httpConnection is possible
    using the uri, apiKey and certificate.
    '''
    http_master_file_path = _nisysmgmt_utils.get_http_master_file()
    if not http_master_file_path or not os.path.exists(http_master_file_path):
        return False

    uri = _nisysmgmt_utils.get_http_master_value('Uri').strip('/') + '/niauth/v1/auth/'
    api_key = _nisysmgmt_utils.get_http_master_value('ApiKey')
    cert_path = _nisysmgmt_utils.get_http_master_value('CertPath')

    try:
        verify = _nisysmgmt_utils.get_http_cert_file() if cert_path else None
        r = requests.get(uri, verify = verify, timeout = 20, headers = {'x-ni-api-key': api_key})
        return r.status_code == 200
    except Exception as e:
        log.debug("Failed to connect to http server %s: %s", uri, e)
        return False


def _update_beacon_interval(interval):
    '''
    Fire an event to update the trigger interval of this beacon
    '''
    _fire_local_event(
        {
        'func': 'modify',
        'name': __virtualname__,
        'beacon_data': [{
            'interval': interval
        }]
        },
        'manage_beacons'
    )


def _http_master_file_found_changed():
    '''
    Check if the http master file status changed
    '''
    global HTTP_MASTER_FILE_WAS_FOUND
    http_master_file_path = _nisysmgmt_utils.get_http_master_file()
    http_master_file_exists = http_master_file_path and os.path.exists(http_master_file_path)
    if(http_master_file_exists != HTTP_MASTER_FILE_WAS_FOUND):
        HTTP_MASTER_FILE_WAS_FOUND = http_master_file_exists
        return True
    return False



def beacon(config):  # pylint: disable=unused-argument
    '''
    Check to see if the http_connected status changed.
    Beacon starts with an interval of 3 sec until the http_master file is found
    and the connection can be checked. Then it switches to a longer, 5 min, interval

    Example Config

    .. code-block:: yaml

       beacons:
         nisysmgmt_http_connected:`
           - interval: 3
    '''
    try:
        if _minion_http_connected_status_changed():
            _fire_status_change_event()
        if _http_master_file_found_changed():
            if HTTP_MASTER_FILE_WAS_FOUND:
                _update_beacon_interval(BEACON_INTERVAL_WITH_HTTP_MASTER_FILE)
            else:
                _update_beacon_interval(BEACON_INTERVAL_WITH_NO_HTTP_MASTER_FILE)
    except Exception as exc:  # pylint: disable=broad-except
        log.error(
            'Unexpected exception in the "nisysmgmt_status_checks" beacon: %s',
            exc, exc_info=True
        )

    # Always return an empty list so that nothing is sent to the master.
    return []
