# -*- coding: utf-8 -*-
'''
National Instruments SystemLink Assets Module
'''
import asyncio
import logging
import os.path
from typing import Any, List, Union, Tuple

from asset_discovery.models import Asset, DiscoveryType, AssetIdentification, ReportingStatus
from asset_discovery.discovery_engine import DiscoveryEngine
from asset_discovery.util.plugin_manager import PluginManager

try:
    # noinspection PyUnresolvedReferences
    from systemlink.messagebus.paths import get_http_master_file
    # noinspection PyUnresolvedReferences
    from systemlink.clientconfig import get_configuration_by_id, HTTP_MASTER_CONFIGURATION_ID
    # noinspection PyUnresolvedReferences
    from systemlink.clientconfig.configuration import Configuration
    # noinspection PyUnresolvedReferences
    from systemlink.clients.niapm import (
        ApiClient, ApiException, AssetsApi,
        AssetUpdateModel, AssetPresenceModel, AssetPresenceStatus,
        CreateAssetsRequest, ExternalCalibrationOperatorModel,
        ExternalCalibrationWithChecksumModel, SelfCalibrationModel,
        TemperatureSensorModel, UpdateAssetsRequest, AssetCreateModel, AssetLocationWithPresenceModel)
    HAS_SYSTEMLINK_SDK = True
except ImportError:
    HAS_SYSTEMLINK_SDK = False

log = logging.getLogger(__name__)

_DISCOVERY_ENGINE = None

__virtualname__ = '_ni_asset_util'


def __virtual__() -> Union[str, Tuple[bool, str]]:
    '''
    During module lazy loading, will return the virtual name of this module.

    :return: The virtual name of this module. On error, return a 2-tuple with
        ``False`` for the first item and the error message for the second.
    '''
    if not HAS_SYSTEMLINK_SDK:
        return False, 'This module requires NI Skyline Message Bus and the python-dateutil module for Python'

    return __virtualname__


def get_asset_discovery_engine():
    global _DISCOVERY_ENGINE  # pylint: disable=global-statement
    if _DISCOVERY_ENGINE:
        return _DISCOVERY_ENGINE
    plugin_manager = PluginManager()
    _DISCOVERY_ENGINE = DiscoveryEngine(plugin_manager)
    return _DISCOVERY_ENGINE


def enable_publishing_on_asset_changes(minion_id: str) -> None:
    '''
    Method for publishing the user defined assets when a change in the asset discovery engine occurs.
    '''
    get_asset_discovery_engine().register_change_callback(
        'publish_assets_callback',
        publish_assets,
        minion_id)


def publish_assets(minion_id: str) -> bool:
    '''
    Function used to retrieve user-specified assets and asset-related information
    from the Data\\Assets\\UserDefined directory on the minion.
    '''
    try:
        assets = get_asset_discovery_engine().list_assets(include_unreported=True)
        asset_updates = _map_assets_to_asset_update_models(assets, minion_id)
        _publish_assets(asset_updates, base_assets=assets)
    except AssetPerformanceManagmentHttpWriterException as exc:
        if exc.is_warning:
            log.warning(exc)
        else:
            log.error(exc)
        return False
    except Exception as exc:  # pylint: disable=broad-except
        log.error(
            'Unexpected exception in "_ni_asset_util module": %s',
            exc, exc_info=True
        )
        return False
    return True


def _map_assets_to_asset_update_models(assets: List[Any], minion_id: str) -> List[AssetUpdateModel]:
    return [_map_asset_to_asset_update_model(asset, minion_id) for asset in assets]


def _map_asset_to_asset_update_model(asset: Any, minion_id: str) -> AssetUpdateModel:
    return AssetUpdateModel(
        model_name=asset.model_name,
        model_number=asset.model_number,
        serial_number=asset.serial_number,
        vendor_name=asset.vendor_name,
        vendor_number=asset.vendor_number,
        bus_type=asset.bus_type,
        name=asset.name,
        asset_type=asset.asset_type,
        firmware_version=asset.firmware_version,
        hardware_version=asset.hardware_version,
        visa_resource_name=asset.visa_resource_name,
        temperature_sensors=_map_temperature_sensors(asset.temperature_sensors),
        supports_self_calibration=asset.supports_self_calibration,
        supports_external_calibration=asset.supports_external_calibration,
        is_ni_asset=asset.is_ni_asset,
        location=_map_location(asset.location, minion_id),
        external_calibration=_map_external_calibration(asset.external_calibration),
        self_calibration=_map_self_calibration(asset.self_calibration),
        properties=asset.properties,
        keywords=asset.keywords
    )


def _map_location(location: Any, minion_id: str) -> Union[AssetLocationWithPresenceModel, None]:
    if location is None:
        return None
    return AssetLocationWithPresenceModel(
        minion_id=minion_id,
        parent=location.parent,
        resource_uri=location.resource_uri,
        slot_number=location.slot_number,
        state=_map_asset_presence(location.state))


def _map_asset_presence(asset_presence: Any) -> Union[AssetPresenceModel, None]:
    if asset_presence is None:
        return None
    return AssetPresenceModel(asset_presence=asset_presence.asset_presence)


def _map_external_calibration(external_calibration: Any) -> Union[ExternalCalibrationWithChecksumModel, None]:
    if external_calibration is None:
        return None
    return ExternalCalibrationWithChecksumModel(
        temperature_sensors=_map_temperature_sensors(external_calibration.temperature_sensors),
        is_limited=external_calibration.is_limited,
        date=external_calibration.date,
        recommended_interval=external_calibration.recommended_interval,
        next_recommended_date=external_calibration.next_recommended_date,
        next_custom_due_date=external_calibration.next_custom_due_date,
        comments=external_calibration.comments,
        entry_type=external_calibration.entry_type,
        operator=_map_operator(external_calibration.operator),
        checksum=external_calibration.checksum)


def _map_self_calibration(self_calibration: Any) -> Union[SelfCalibrationModel, None]:
    if self_calibration is None:
        return None
    return SelfCalibrationModel(
        temperature_sensors=_map_temperature_sensors(self_calibration.temperature_sensors),
        is_limited=self_calibration.is_limited,
        date=self_calibration.date)


def _map_temperature_sensors(temperature_sensors: List[Any]) -> List[TemperatureSensorModel]:
    return [TemperatureSensorModel(name=sensor.name, reading=sensor.reading) for sensor in temperature_sensors]


def _map_operator(operator: Any) -> ExternalCalibrationOperatorModel:
    return ExternalCalibrationOperatorModel(display_name=operator.display_name, user_id=operator.user_id)


def _publish_assets(
        asset_update_models: List[AssetUpdateModel],
        base_assets=None) -> None:
    '''
    Call the writer to publish the assets over HTTP.

    :param asset_update_models: A list of AssetUpdateModel objects representing the assets.
    '''
    if not asset_update_models:
        return

    loop = None
    new_event_loop_created = None
    try:
        new_event_loop_created = _ensure_event_loop()
        loop = asyncio.get_event_loop()
        loop.run_until_complete(
            _update_assets(
                asset_update_models,
                base_assets
            )
        )
    finally:
        if new_event_loop_created is True and loop is not None:
            loop.stop()
            loop.close()
            asyncio.set_event_loop(None)


def _ensure_event_loop() -> bool:
    try:
        asyncio.get_event_loop()
        return False
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        return True


async def _update_assets(
        asset_update_models,
        base_assets) -> None:
    '''
    Call the writer to publish the assets over HTTP in an async manner.

    :param asset_update_models: A list of AssetUpdateModel objects representing the assets.
    '''
    writer = AssetPerformanceManagmentHttpWriter()
    if writer.get_configuration() is None:
        log.warning("Failed to create the AssetPerformanceManagmentHttpWriter with a valid config.")
        return
    not_found_assets = await writer.update_assets(asset_update_models)
    if not_found_assets:
        not_found_assets = _filter_out_unreported_models(not_found_assets)
        asset_create_models = _convert_update_to_create_models(not_found_assets, base_assets)
        await writer.create_assets(asset_create_models)


def _filter_out_unreported_models(not_found_assets):
    discovery_engine = get_asset_discovery_engine()
    asset_metadata = discovery_engine.get_metadata()
    if asset_metadata.status.succeeded is False:
        log.error("Failed to retrieve asset metadata.")
        return []
    asset_metadata = asset_metadata.content
    asset_metadata_dict = {m.asset_identification: m for m in asset_metadata}
    reported_not_found_assets = []
    for asset in not_found_assets:
        identification = AssetIdentification(vendor_name=asset.vendor_name,
                                             vendor_number=asset.vendor_number,
                                             model_name=asset.model_name,
                                             model_number=asset.model_number,
                                             serial_number=asset.serial_number)
        metadata_for_asset = asset_metadata_dict[identification]
        if metadata_for_asset.reporting_status != ReportingStatus.UNREPORTED_RESOURCE:
            reported_not_found_assets.append(asset)
    return reported_not_found_assets


def _convert_update_to_create_models(  # pylint: disable=invalid-name
        asset_update_models: List[AssetUpdateModel],
        base_assets: List[Asset]) -> List[AssetCreateModel]:
    '''
    Converts a list of AssetUpdateModel objects into AssetCreateModel objects.

    :param asset_update_models: The input list of AssetUpdateModel objects.
    :return: The converted list of AssetCreateModel objects.
    '''
    return [
        _convert_update_to_create_model(update_model, _get_discovery_type_for_asset(update_model, base_assets))
        for update_model in asset_update_models
    ]


def _get_discovery_type_for_asset(
        update_model: AssetUpdateModel,
        base_assets: List[Asset]) -> Union[DiscoveryType, None]:
    if base_assets is None:
        return None

    for base_asset in base_assets:
        if base_asset.serial_number == update_model.serial_number \
                and base_asset.model_name == update_model.model_name \
                and base_asset.model_number == update_model.model_number \
                and base_asset.vendor_name == update_model.vendor_name \
                and base_asset.vendor_number == update_model.vendor_number:
            return base_asset.discovery_type
    return None


def _convert_update_to_create_model(
        asset_update_model: AssetUpdateModel,
        discovery_type: DiscoveryType) -> AssetCreateModel:
    return AssetCreateModel(
            model_name=asset_update_model.model_name,
            model_number=asset_update_model.model_number,
            serial_number=asset_update_model.serial_number,
            vendor_name=asset_update_model.vendor_name,
            vendor_number=asset_update_model.vendor_number,
            bus_type=asset_update_model.bus_type,
            name=asset_update_model.name,
            asset_type=asset_update_model.asset_type,
            firmware_version=asset_update_model.firmware_version,
            hardware_version=asset_update_model.hardware_version,
            visa_resource_name=asset_update_model.visa_resource_name,
            temperature_sensors=asset_update_model.temperature_sensors,
            supports_self_calibration=asset_update_model.supports_self_calibration,
            supports_external_calibration=asset_update_model.supports_external_calibration,
            self_calibration=asset_update_model.self_calibration,
            is_ni_asset=asset_update_model.is_ni_asset,
            location=asset_update_model.location,
            external_calibration=asset_update_model.external_calibration,
            keywords=asset_update_model.keywords,
            properties=asset_update_model.properties,
            discovery_type=discovery_type
        )


class AssetPerformanceManagmentHttpWriterException(Exception):
    '''
    Exception raised by the AssetPerformanceManagmentHttpWriter class.
    '''

    def __init__(
            self,
            message: str,
            inner_exception: Exception = None,
            is_warning: bool = None) -> None:
        '''
        Initialize the exception.

        :param message: The message of the exception.
        :param inner_exception: The inner exception of the exception.
        :param is_warning: Whether the exception is a warning.
        '''
        if inner_exception:
            msg = '{0}: {1}'.format(message, inner_exception)
        else:
            msg = message

        super().__init__(msg)

        self.is_warning = is_warning is True


class AssetPerformanceManagmentHttpWriter:  # pylint: disable=too-few-public-methods
    '''
    Abstraction over HTTP connection which facilitates communication
    between the minion and the AssetPerformanceManagement service.
    '''

    def __init__(self) -> None:
        '''
        Initialize the message service if not already initialized
        '''
        self._configuration = None
        self._setup_communication()

    def get_configuration(self) -> Configuration:
        return self._configuration

    def _setup_communication(self) -> None:
        '''
        Set up the communication.
        '''
        file_path = get_http_master_file()
        if not os.path.isfile(file_path):
            # The Skyline Master file is not available.
            # Can't set up the Message Service without it.
            return

        self._setup_http_connection()

    def _setup_http_connection(self) -> None:
        '''
        Set up the HTTP communication and initialize the Assets API object.
        '''
        assets_route_name = '/niapm'
        self._configuration = get_configuration_by_id(
            HTTP_MASTER_CONFIGURATION_ID,
            assets_route_name,
            False)

    async def update_assets(
            self,
            asset_update_models: List[AssetUpdateModel]) -> List[AssetUpdateModel]:
        '''
        Publish an /update-assets request over HTTP.

        :param asset_update_models: A list of AssetUpdateModel objects representing the assets.
        :return: A list of AssetUpdateModels objects that were not found on the server.
        '''
        try:
            not_found_error_code = -254772

            update_assets_request = UpdateAssetsRequest(
                asset_update_models
            )

            async with ApiClient(self._configuration) as api_client:
                assets_api = AssetsApi(api_client=api_client)
                update_assets_response = await assets_api.update_assets(
                    update_assets_request)

                failed_assets = update_assets_response.failed
                error = update_assets_response.error
                if error is None:
                    return []
                inner_errors = error.inner_errors
                if inner_errors is None:
                    return []
                not_found_assets = []
                for i, inner_error in enumerate(inner_errors):
                    if inner_error.code == not_found_error_code:
                        not_found_assets.append(failed_assets[i])
                return not_found_assets
        except ApiException as exc:
            if exc.status == 401:
                # The salt-master may have changed credentials. Notify this accordingly.
                raise AssetPerformanceManagmentHttpWriterException(
                    'HTTP Authentication error. Credentials may have changed',
                    exc
                )
            # All other HTTP exceptions. Notify this accordingly.
            raise AssetPerformanceManagmentHttpWriterException(
                'An HTTP error has occurred',
                exc
            )

    async def create_assets(self, asset_create_models: List[AssetCreateModel]) -> None:
        '''
        Publish a /create-assets request over HTTP.

        :param asset_create_models: A list of AssetCreateModel objects representing the assets.
        '''
        try:
            create_assets_request = CreateAssetsRequest(
                asset_create_models
            )

            async with ApiClient(self._configuration) as api_client:
                assets_api = AssetsApi(api_client=api_client)
                await assets_api.create_assets(
                    create_assets_request)
        except ApiException as exc:
            if exc.status == 401:
                # The salt-master may have changed credentials. Notify this accordingly.
                raise AssetPerformanceManagmentHttpWriterException(
                    'HTTP Authentication error. Credentials may have changed',
                    exc
                )
            # All other HTTP exceptions. Notify this accordingly.
            raise AssetPerformanceManagmentHttpWriterException(
                'An HTTP error has occurred',
                exc
            )
