# coding: utf-8
'''
National Instruments SystemLink Assets Module
'''
import _ni_asset_util as ni_asset_util
from asset_discovery.models import (
    Asset,
    AssetUpdate,
    AssetIdentification,
    TemperatureSensor,
    SelfCalibration,
    Location,
    LocationUpdate,
    AssetPresence,
    ExternalCalibration,
    CalibrationOperator,
    Defaults)
from asset_discovery.util.asset_serializer import AssetSerializer

__virtualname__ = 'ni_asset'

DISCOVERY_ENGINE = None
ASSET_SERIALIZER = None
ASSET_UPDATE_SERIALIZER = None
DEFAULT = Defaults.DEFAULT_VALUE


def __virtual__():
    global DISCOVERY_ENGINE  # pylint: disable=global-statement
    global ASSET_SERIALIZER  # pylint: disable=global-statement
    global ASSET_UPDATE_SERIALIZER  # pylint: disable=global-statement
    DISCOVERY_ENGINE = ni_asset_util.get_asset_discovery_engine()
    ASSET_SERIALIZER = AssetSerializer()
    ASSET_UPDATE_SERIALIZER = AssetSerializer(infer_present=False)
    return __virtualname__


def list_assets(**kwargs):
    '''
    Lists all assets.
    '''
    assets = DISCOVERY_ENGINE.list_assets(**kwargs)
    return [a.to_dict() for a in assets]


def add_asset(name=None,
              serial_number=None,
              model_name=None,
              model_number=None,
              vendor_name=None,
              vendor_number=None,
              bus_type=None,
              asset_type=None,
              firmware_version=None,
              hardware_version=None,
              visa_resource_name=None,
              temperature_sensors=None,
              supports_self_calibration=None,
              supports_external_calibration=None,
              self_calibration=None,
              is_ni_asset=None,
              location=None,
              external_calibration=None,
              keywords=None,
              properties=None,
              **kwargs):
    '''
    Adds an asset.
    '''
    asset_props = _map_asset_props(locals())
    asset = ASSET_SERIALIZER.from_json(asset_props)
    if asset is None:
        raise Exception("Invalid asset")
    DISCOVERY_ENGINE.add_asset(asset, **kwargs)
    return True


def _map_asset_props(asset):
    keys = Asset.serialization_keys
    asset_props = {
        keys["model_name"]: asset.get("model_name", None),  # The identification props are not inlined on update
        keys["model_number"]: asset.get("model_number", None),
        keys["serial_number"]: asset.get("serial_number", None),
        keys["vendor_name"]: asset.get("vendor_name", None),
        keys["vendor_number"]: asset.get("vendor_number", None),
        keys["bus_type"]: asset["bus_type"],
        keys["name"]: asset["name"],
        keys["asset_type"]: asset["asset_type"],
        keys["firmware_version"]: asset["firmware_version"],
        keys["hardware_version"]: asset["hardware_version"],
        keys["visa_resource_name"]: asset["visa_resource_name"],
        keys["temperature_sensors"]: _map_temperature_sensors_props(asset["temperature_sensors"]),
        keys["supports_self_calibration"]: asset["supports_self_calibration"],
        keys["supports_external_calibration"]: asset["supports_external_calibration"],
        keys["self_calibration"]: _map_self_calibration_props(asset["self_calibration"]),
        keys["is_ni_asset"]: asset["is_ni_asset"],
        keys["location"]: _map_location_props(asset["location"]),
        keys["external_calibration"]: _map_external_calibration_props(asset["external_calibration"]),
        keys["keywords"]: asset["keywords"],
        keys["properties"]: asset["properties"]
    }
    return _filter_out_default_properties(asset_props)


def _filter_out_default_properties(properties):
    return {k: v for k, v in properties.items() if v is not DEFAULT}


def _map_temperature_sensors_props(temperature_sensors):
    if temperature_sensors is None or temperature_sensors is DEFAULT:
        return temperature_sensors

    return [_map_temperature_sensor_props(temperature_sensor) for temperature_sensor in temperature_sensors]


def _map_temperature_sensor_props(temperature_sensor):
    if temperature_sensor is None:
        return None

    keys = TemperatureSensor.serialization_keys
    return {
        keys["name"]: temperature_sensor.get("name", None),
        keys["reading"]: temperature_sensor["reading"]  # Required
    }


def _map_self_calibration_props(self_calibration):
    if self_calibration is None or self_calibration is DEFAULT:
        return self_calibration

    keys = SelfCalibration.serialization_keys
    return {
        keys["temperature_sensors"]: _map_temperature_sensors_props(self_calibration.get("temperature_sensors", None)),
        keys["is_limited"]: self_calibration.get("is_limited", None),
        keys["date"]: self_calibration["date"]  # Required
    }


def _map_location_props(location):
    if location is None or location is DEFAULT:
        return location

    keys = Location.serialization_keys
    location_props = {
        keys["parent"]: location.get("parent", DEFAULT),
        keys["resource_uri"]: location.get("resource_uri", DEFAULT),
        keys["slot_number"]: location.get("slot_number", DEFAULT),
        keys["state"]: _map_state_props(location.get("state", DEFAULT))
    }
    return _filter_out_default_properties(location_props)


def _map_state_props(state):
    if state is None or state is DEFAULT:
        return state

    keys = AssetPresence.serialization_keys
    return {
        keys["asset_presence"]: state["asset_presence"]  # Required
    }


def _map_external_calibration_props(external_calibration):
    if external_calibration is None or external_calibration is DEFAULT:
        return external_calibration

    keys = ExternalCalibration.serialization_keys
    return {
        # Required
        keys["temperature_sensors"]: _map_temperature_sensors_props(external_calibration["temperature_sensors"]),
        keys["is_limited"]: external_calibration.get("is_limited", None),
        keys["date"]: external_calibration["date"],  # Required
        keys["recommended_interval"]: external_calibration["recommended_interval"],  # Required
        keys["next_recommended_date"]: external_calibration.get("next_recommended_date", None),
        keys["next_custom_due_date"]: external_calibration.get("next_custom_due_date", None),
        keys["comments"]: external_calibration.get("comments", None),
        keys["entry_type"]: external_calibration.get("entry_type", None),
        keys["operator"]: _map_operator_props(external_calibration.get("operator", None)),
        keys["checksum"]: external_calibration.get("checksum", None)
    }


def _map_operator_props(operator):
    if operator is None:
        return None

    keys = CalibrationOperator.serialization_keys
    return {
        keys["display_name"]: operator.get("display_name", None),
        keys["user_id"]: operator.get("user_id", None)
    }


def remove_asset(name=None, **kwargs):
    '''
    Removes an asset.
    '''
    identification_props = kwargs.get("identification", None)
    asset = _get_single_asset_by_name(name, identification_props)

    DISCOVERY_ENGINE.remove_asset(asset.identification, **kwargs)
    return True


def _get_single_asset_by_name(name, identification_props):
    assets = DISCOVERY_ENGINE.list_assets()

    # Try to get the asset by name
    assets_matching_name = [a for a in assets if a.name == name]

    # If we found a single asset with the given name, use that
    if len(assets_matching_name) == 1:
        return assets_matching_name[0]

    # If we found none or more than one asset with the given name,
    # try to get the asset by identification_props
    identification = _create_asset_identification(identification_props)
    assets_to_query_by_id = assets if len(assets_matching_name) == 0 else assets_matching_name
    assets_matching_id = [a for a in assets_to_query_by_id if a.identification == identification]

    # If we found a single asset with the given identification_props, use that
    if len(assets_matching_id) == 1:
        return assets_matching_id[0]

    # We either found no resources or a multiple resources, so we can't perform the operation
    raise Exception("Could not find a single asset based on name and optional `identification` kwarg")


def _create_asset_identification(identification_props):
    if isinstance(identification_props, dict):
        return AssetIdentification(
            model_name=identification_props.get("model_name", None),
            model_number=identification_props.get("model_number", None),
            serial_number=identification_props.get("serial_number", None),
            vendor_name=identification_props.get("vendor_name", None),
            vendor_number=identification_props.get("vendor_number", None)
        )
    return None


def update_asset(name=DEFAULT,
                 updated_name=DEFAULT,
                 bus_type=DEFAULT,
                 asset_type=DEFAULT,
                 firmware_version=DEFAULT,
                 hardware_version=DEFAULT,
                 visa_resource_name=DEFAULT,
                 temperature_sensors=DEFAULT,
                 supports_self_calibration=DEFAULT,
                 supports_external_calibration=DEFAULT,
                 self_calibration=DEFAULT,
                 is_ni_asset=DEFAULT,
                 location=DEFAULT,
                 external_calibration=DEFAULT,
                 keywords=DEFAULT,
                 properties=DEFAULT,
                 **kwargs):
    '''
    Updates an asset.
    '''
    asset_props = _map_asset_props(locals())
    identification_props = kwargs.get("identification", None)
    asset = _get_single_asset_by_name(name, identification_props)

    asset_update = _create_asset_update(asset, updated_name, asset_props)
    if asset_update is None:
        raise Exception("Invalid asset")

    DISCOVERY_ENGINE.update_asset(asset_update, **kwargs)
    return True


def _create_asset_update(asset, updated_name, asset_props):
    asset_keys = Asset.serialization_keys
    asset_props[asset_keys["model_name"]] = asset.model_name
    asset_props[asset_keys["model_number"]] = asset.model_number
    asset_props[asset_keys["serial_number"]] = asset.serial_number
    asset_props[asset_keys["vendor_name"]] = asset.vendor_name
    asset_props[asset_keys["vendor_number"]] = asset.vendor_number
    asset_props[asset_keys["name"]] = updated_name
    parsed_asset = ASSET_UPDATE_SERIALIZER.from_json(asset_props)

    if parsed_asset is None:
        return None

    asset_update = AssetUpdate.create_from_identification(parsed_asset.identification)
    asset_update.location = _get_updated_location(asset_keys["location"], asset_props, parsed_asset.location)
    asset_update.bus_type = _get_updated_value(asset_keys["bus_type"], asset_props, parsed_asset.bus_type)
    asset_update.name = _get_updated_value(asset_keys["name"], asset_props, parsed_asset.name)
    asset_update.asset_type = _get_updated_value(asset_keys["asset_type"], asset_props, parsed_asset.asset_type)
    asset_update.firmware_version = _get_updated_value(asset_keys["firmware_version"], asset_props, parsed_asset.firmware_version)
    asset_update.hardware_version = _get_updated_value(asset_keys["hardware_version"], asset_props, parsed_asset.hardware_version)
    asset_update.visa_resource_name = _get_updated_value(asset_keys["visa_resource_name"], asset_props, parsed_asset.visa_resource_name)
    asset_update.temperature_sensors = _get_updated_value(asset_keys["temperature_sensors"], asset_props, parsed_asset.temperature_sensors)
    asset_update.supports_self_calibration = _get_updated_value(asset_keys["supports_self_calibration"], asset_props, parsed_asset.supports_self_calibration)
    asset_update.supports_external_calibration = _get_updated_value(asset_keys["supports_external_calibration"], asset_props, parsed_asset.supports_external_calibration)
    asset_update.self_calibration = _get_updated_value(asset_keys["self_calibration"], asset_props, parsed_asset.self_calibration)
    asset_update.is_ni_asset = _get_updated_value(asset_keys["is_ni_asset"], asset_props, parsed_asset.is_ni_asset)
    asset_update.external_calibration = _get_updated_value(asset_keys["external_calibration"], asset_props, parsed_asset.external_calibration)
    asset_update.properties = _get_updated_value(asset_keys["properties"], asset_props, parsed_asset.properties)
    asset_update.keywords = _get_updated_value(asset_keys["keywords"], asset_props, parsed_asset.keywords)

    return asset_update


def _get_updated_location(location_key, asset_props, parsed_location):
    location_keys = Location.serialization_keys
    if location_key in asset_props:
        location_props = asset_props[location_key]
        if parsed_location is None:
            if location_props is None:
                return None
            else:
                return DEFAULT
        else:
            location_update = LocationUpdate()
            location_update.parent = _get_updated_value(location_keys["parent"], location_props, parsed_location.parent)
            location_update.resource_uri = _get_updated_value(location_keys["resource_uri"], location_props, parsed_location.resource_uri)
            location_update.slot_number = _get_updated_value(location_keys["slot_number"], location_props, parsed_location.slot_number)
            location_update.state = _get_updated_value(location_keys["state"], location_props, parsed_location.state)
            return location_update
    return DEFAULT


def _get_updated_value(key, raw_properties, parsed_value):
    '''
    Based on existence of a key in a dictionary, this function returns the correct value for a property.

    It covers the following scenarios:
    - If the key exists in the dictionary, the user specified some value for it
        And, if the parsed value is None matching a None value in the dictionary,
            then it was user intention to set it to None, so we return None.
        However, if the parsed value is None not matching a non-None value in the dictionary,
            then an invalid value was specified for that property, so we return DEFAULT (no change).
        Alternatively, if the parsed value is something other than None,
            then a valid values was specified for that property, se we return it.
    - If the key doesn't exist in the dictionary, the user did not specify any value for it
        So, we return DEFAULT (no change).
    '''
    if key in raw_properties:
        value = raw_properties[key]
        if parsed_value is None:
            if value is None:
                return None
            else:
                return DEFAULT
        else:
            return parsed_value
    return DEFAULT


def rename_asset(name=None, updated_name=None, **kwargs):
    '''
    Renames an asset.
    '''
    identification_props = kwargs.get("identification", None)
    asset = _get_single_asset_by_name(name, identification_props)

    asset_update_with_name = _create_asset_update_with_name(asset, updated_name)
    DISCOVERY_ENGINE.update_asset(asset_update_with_name, **kwargs)
    return True


def _create_asset_update_with_name(asset, updated_name):
    asset_update = AssetUpdate.create_from_identification(asset.identification)
    asset_update.name = updated_name
    return asset_update


def refresh(**kwargs):
    '''
    Performs a refresh on all assets if kwargs is empty or on the assets specified by kwargs.
    '''
    refresh_result = DISCOVERY_ENGINE.refresh(**kwargs)
    if refresh_result != 0:
        # log this
        return []
    refreshed_assets = DISCOVERY_ENGINE.list_assets(**kwargs)
    return [a.to_dict() for a in refreshed_assets]


def reset(name=None, **kwargs):
    """
    Executes a reset operation on an asset with the given name.
    If multiple assets match the name the operation fails.

    The `identification` kwarg is used so the caller can more specifically
    identify the resource (in addition or in lieu of the name).

    The `mode` kwarg can be used to specify the mode of the operation.

    Returns True/False for success/failure.
    """
    identification_props = kwargs.get("identification", None)
    asset = _get_single_asset_by_name(name, identification_props)

    reset_result = DISCOVERY_ENGINE.reset(asset.identification, **kwargs)

    return reset_result


def self_test(name=None, **kwargs):
    '''
    Executes a self test operation on the asset with the given name.
    If multiple assets match the given name the operation fails.

    Callers can use an `identification` kwarg to be more specific
    about the resource to operate on.

    The `mode` kwarg can be used to specify the mode of the operation.

    Returns True/False for success/failure.
    '''
    identification_props = kwargs.get('identification', None)
    asset = _get_single_asset_by_name(name, identification_props)

    return DISCOVERY_ENGINE.self_test(asset.identification, **kwargs)


def self_calibrate(name=None, **kwargs):
    '''
    Executes a self test operation on the asset with the given name.
    If multiple assets match the given name the operation fails.

    Callers can use an `identification` kwarg to be more specific
    about the resource to operate on.

    Returns True/False for success/failure.
    '''
    identification_props = kwargs.get('identification', None)
    asset = _get_single_asset_by_name(name, identification_props)

    return DISCOVERY_ENGINE.self_calibrate(asset.identification, **kwargs)
